/**
 * [NOTE] The GUI code is very sloppy, especially the font-sensitivity stuff. 
 *
 * I really wish Be would incorporate font-sensitivity in the Interface Kit.
 * Or maybe I should just learn to use a GUI builder tool;-)
 */

/**
 * [NOTE] SavePrefs() is called many times when the user moves or resizes the
 * window. I could prevent this by checking whether the mouse button is up
 * or down, and only save the settings if it is up.
 */
  
// ***************************************************************************
// Implementation System Includes
// ***************************************************************************

// ***************************************************************************
// Implementation Project Includes
// ***************************************************************************

#include "Main.h"
#include "GrepWindow.h"
#include "Version.h"
#include "Fit.h"

// ***************************************************************************
// Implementation Constant Definitions
// ***************************************************************************

/** The messages that its views send to the window. */
enum {
  /** Start the grepping. */
  MSG_START_SEARCH = 1000,

  /** Cancel the grepping. */
  MSG_CANCEL_SEARCH,

  /** The value of the "links" check box has changed. */
  MSG_LINKS_CHECKBOX_TOGGLED,
  
  /** The value of the "subdirectories" check box has changed. */
  MSG_SUBDIR_CHECKBOX_TOGGLED,

  /** The value of the "show contents" check box has changed. */
  MSG_SHOWCONTENTS_CHECKBOX_TOGGLED,
  
  // dble-click a result
  MSG_INVOKE_ITEM
};

const tChar* PREFS_FILE = "/boot/home/config/settings/TrackerGrepSettings";

// The initial dimensions of the window.
const float WINDOW_X      = 100;
const float WINDOW_Y      = 100;
const float WINDOW_WIDTH  = 420;
const float WINDOW_HEIGHT = 300;

// The minimum and maximum dimensions of the window.
const float WINDOW_MIN_WIDTH  = WINDOW_WIDTH;
const float WINDOW_MIN_HEIGHT = 150;
const float WINDOW_MAX_WIDTH  = 100000; // This should be large enough.
const float WINDOW_MAX_HEIGHT = 100000; // Anybody got a 40" monitor :-)

class BEntryItem : public BStringItem {
public:
	BEntryItem(entry_ref *ref):BStringItem("",0,false)
	{
		fRef=*ref;
		BEntry entry(ref);
		BPath path(&entry);
		SetText(path.Path()); 
	}
	
	entry_ref *Ref() {return &fRef;}
	
private:
	entry_ref fRef;
};
// ***************************************************************************
// Implementation Type Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Variable Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Variable Definitions
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Lifecycle
// ***************************************************************************

// ===========================================================================
// Constructor
// ===========================================================================

   
tWindow::tWindow(entry_ref aDirRef, BMessage* apMessage) 
  : BWindow(Goowee_Fit(BRect(WINDOW_X, 
                             WINDOW_Y, 
                             WINDOW_X + WINDOW_WIDTH,
                             WINDOW_Y + WINDOW_HEIGHT)),
            "Tracker Grep",
            B_DOCUMENT_WINDOW,
            0)
{
  // _KMSF_012099_
  /*
   	the grepper object is now local to the window, not global...
   	it's better suited if we want to launch multiple-copies of TG without SegFault
  */

  mpGrepper=new tGrepper(aDirRef,apMessage);

  // Create the background box.
  BBox* lpBox  = new BBox(Bounds(),
          B_EMPTY_STRING, 
          B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP, 
          B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE_JUMP, 
          B_PLAIN_BORDER); 
          
  AddChild(lpBox);

  // Create the search pattern's entry field. This view does not send a 
  // message when <Enter> is pressed, because the "search/cancel" button
  // already does this (we don't want to send the same message twice.)
  mpSearchText = new BTextControl(BRect(8,
                                        8,
                                        Goowee_Fit(WINDOW_WIDTH)-8,
                                        1),
          "SearchText",
          0,                           // No label.
          0,                           // No initial text.
          0,                           // No message.
          B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP,
          B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE | B_NAVIGABLE); 

  mpSearchText->ResizeToPreferred();   // Because it doesn't have a label.
  lpBox->AddChild(mpSearchText);
  mpSearchText->MakeFocus(TRUE);       // Must come after AddChild().
  mpSearchText->TextView()->SetMaxBytes(1000); 

  // Create the "search/cancel" button
  float lTop = mpSearchText->Frame().bottom+8;
  mpButton = new BButton(BRect(Goowee_Fit(WINDOW_WIDTH-80)-12,
                               lTop,
                               Goowee_Fit(WINDOW_WIDTH)-12,
                               1),
          "Button",
          "Search",
          new BMessage(MSG_START_SEARCH),
          B_FOLLOW_RIGHT);

  mpButton->MakeDefault(TRUE);          
  lpBox->AddChild(mpButton); 
  lpBox->ResizeTo(Goowee_Fit(WINDOW_WIDTH),
                  19 + mpSearchText->Bounds().Height()
                     + mpButton->Bounds().Height());

  // Create the "links" check box.
  lTop += Goowee_Fit(3);
  mpRecurseLinks = new BCheckBox(BRect(8,
                                       lTop,
                                       Goowee_Fit(70),
                                       1),
          "RecurseLinks",
          "Links",
          new BMessage(MSG_LINKS_CHECKBOX_TOGGLED),
          B_FOLLOW_LEFT);
          
  mpRecurseLinks->SetValue(B_CONTROL_OFF);
  lpBox->AddChild(mpRecurseLinks);

  // Create the "subdirectories" check box.
  mpRecurseDirs = new BCheckBox(BRect(Goowee_Fit(80),
                                      lTop,
                                      Goowee_Fit(190),
                                      1),
          "RecurseDirs",
          "Subdirectories",
          new BMessage(MSG_SUBDIR_CHECKBOX_TOGGLED),
          B_FOLLOW_LEFT);
          
  mpRecurseDirs->SetValue(B_CONTROL_ON);
  lpBox->AddChild(mpRecurseDirs);

  // Create the "show contents" check box.
  mpShowContents = new BCheckBox(BRect(Goowee_Fit(200),
                                      lTop,
                                      Goowee_Fit(310),
                                      1),
          "ShowContents",
          "Show contents",
          new BMessage(MSG_SHOWCONTENTS_CHECKBOX_TOGGLED),
          B_FOLLOW_LEFT);
          
  mpShowContents->SetValue(B_CONTROL_ON);
  lpBox->AddChild(mpShowContents);
  
  // Create the search results' display.
  BRect lRect = Bounds();
  lRect.top = lpBox->Frame().bottom+1;
/*
  mpSearchResults = new BTextView(lRect,
          "SearchResult", 
          BRect(0,0,10000,10000),      // Always enough space to display text.
          B_FOLLOW_ALL_SIDES,
          B_WILL_DRAW);

  mpSearchResults->MakeEditable(FALSE);
  mpSearchResults->MakeSelectable(FALSE);
  mpSearchResults->SetStylable(TRUE);
  mpSearchResults->SetWordWrap(FALSE);
  mpSearchResults->SetAutoindent(FALSE);
*/
  mpSearchResults = new BOutlineListView(lRect,
  		"SearchResult", 
  		B_SINGLE_SELECTION_LIST,
		B_FOLLOW_ALL_SIDES,
        B_WILL_DRAW);
  // Write some default text.
  //mpSearchResults->SetFontAndColor(be_bold_font);
  //mpSearchResults->Insert("Please enter something to search for.");
  mpSearchResults->SetInvocationMessage(new BMessage(MSG_INVOKE_ITEM));

  // Connect the grepper object to this window.
  mpGrepper->SetTarget(this);
  
  // Put scroll bars around the search results.  
  BScrollView* lpScroller = new BScrollView("ScrollSearchResults", 
          mpSearchResults, 
          B_FOLLOW_ALL_SIDES, 
          B_FULL_UPDATE_ON_RESIZE, 
          TRUE,                        // Horizontal scrollbar.
          TRUE,                        // Vertical scrollbar.
          B_NO_BORDER);
          
  AddChild(lpScroller);
  lpScroller->ResizeTo(Goowee_Fit(WINDOW_WIDTH)+1,
                       lRect.Height()+1);

  // Adjust size limits of the window.  
  SetSizeLimits(Goowee_Fit(WINDOW_MIN_WIDTH), 
                Goowee_Fit(WINDOW_MAX_WIDTH),
                Goowee_Fit(WINDOW_MIN_HEIGHT),
                Goowee_Fit(WINDOW_MAX_HEIGHT));

  // Adjust the new window's settings to the last saved preferences.
  LoadPrefs();                
  Show(); 
}

// ===========================================================================
// Destructor
// ===========================================================================

tWindow::~tWindow()
{
	delete mpGrepper;
}
     
// ***************************************************************************
// PUBLIC Member Function Definitions: Operators
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Operations
// ***************************************************************************

// ===========================================================================
// MessageReceived
// ===========================================================================

void
tWindow::MessageReceived(BMessage* apMessage)
{
  // Figure out what kind of message was received.
  switch(apMessage->what) {
  
    // We need to start a new search. ----------------------------------------
    case MSG_START_SEARCH: {

      // Quit the current grepper thread (if there is one).
      mpGrepper->Stop();
        
      // Clear the search results view.
      //mpSearchResults->Delete(0,mpSearchResults->TextLength());
      mpSearchResults->MakeEmpty();
    
      // Only start the search if the user actually entered some text.
      if (0 == strcmp(mpSearchText->Text(),"")) {
        //mpSearchResults->SetFontAndColor(be_bold_font);
        //mpSearchResults->Insert("Please enter something to search for.");

      // Let's grep.
      } else {
        mpGrepper->SetPattern(mpSearchText->Text());
        if (B_OK == mpGrepper->Start()) {
          mpRecurseDirs->SetEnabled(FALSE);
          mpRecurseLinks->SetEnabled(FALSE);
          mpShowContents->SetEnabled(FALSE);
          mpSearchText->SetEnabled(FALSE);
          mpButton->MakeFocus(TRUE);
          mpButton->SetLabel("Cancel");
          mpButton->SetMessage(new BMessage(MSG_CANCEL_SEARCH));
          
          // Remember the search pattern.
          strcpy(mSearchPattern,mpSearchText->Text());
        }
      }
      break;
    }

    // We need to cancel the search. -----------------------------------------
    case MSG_CANCEL_SEARCH: {
      mpGrepper->Stop();

      // Fall through.
    }
    
    // The grepper object has finished searching. ----------------------------
    case MSG_SEARCH_FINISHED: {
      mpRecurseDirs->SetEnabled(TRUE);
      mpRecurseLinks->SetEnabled(TRUE);
      mpShowContents->SetEnabled(TRUE);
      mpSearchText->SetEnabled(TRUE);
      mpSearchText->MakeFocus(TRUE);
      mpButton->SetLabel("Search");
      mpButton->SetMessage(new BMessage(MSG_START_SEARCH));

      // Put the search pattern back into the text control.
      mpSearchText->SetText(mSearchPattern);
      mpSearchText->TextView()->SelectAll();
      break;
    }
    
    // The grepper object sends us a file name. ------------------------------
    case MSG_REPORT_FILE_NAME: {

      tChar lTempString[MAX_LEN_REPORT_TEXT];
      strcpy(lTempString,apMessage->FindString("filename"));
      // printf("[Tracker Grep] %s\n",lTempString);

      mpSearchText->SetText(lTempString);
      break;
    }
    
    // The grepper object sends us the output of a grep operation. -----------
    case MSG_REPORT_RESULT: {

		const char *buf;
		entry_ref ref;
		BListItem *item = 0;
		if (apMessage->FindRef("ref",&ref)==B_OK) {
			mpSearchResults->AddItem(item=new BEntryItem(&ref));
  	    }

  	    if (B_CONTROL_ON == mpShowContents->Value()) {
  		  item->SetExpanded(true);
  		} else {
  		  item->SetExpanded(false);
  		}

		 type_code type;
		 int32 count;
		 apMessage->GetInfo("text",&type,&count);
		 while (apMessage->FindString("text",--count,&buf)==B_OK) {
			mpSearchResults->AddUnder(new BStringItem(buf),item);
		 }
		 
      break;
    }

    // The user clicked on the "links" check box. ----------------------------
    case MSG_LINKS_CHECKBOX_TOGGLED: {

      if (B_CONTROL_ON == mpRecurseLinks->Value()) {
        mpGrepper->SetRecurseLinks(TRUE);
      } else {
        mpGrepper->SetRecurseLinks(FALSE);
      }
      SavePrefs();
      break;
    }
        
    // The user clicked on the "subdirectories" check box. -------------------
    case MSG_SUBDIR_CHECKBOX_TOGGLED: {

      if (B_CONTROL_ON == mpRecurseDirs->Value()) {
        mpGrepper->SetRecurseDirs(TRUE);
      } else {
        mpGrepper->SetRecurseDirs(FALSE);
      }
      SavePrefs();
      break;
    }

    // The user clicked on the "show contents" check box. --------------------
    case MSG_SHOWCONTENTS_CHECKBOX_TOGGLED: {
 
      // Loop through the list's items and expand or collapse them all.
      tInt32 lNumItems = mpSearchResults->FullListCountItems();
      if (B_CONTROL_ON == mpShowContents->Value()) {
        for (tInt32 lIndex=0; lIndex<lNumItems; ++lIndex) {
          mpSearchResults->Expand(mpSearchResults->FullListItemAt(lIndex));
        }
      } else {
        for (tInt32 lIndex=0; lIndex<lNumItems; ++lIndex) {
          mpSearchResults->Collapse(mpSearchResults->FullListItemAt(lIndex));
        }
      }
      SavePrefs();
      break;
    }
    
    // The user double-clicked on a result. ----------------------------------
    case MSG_INVOKE_ITEM: {
    	BListItem *item=mpSearchResults->ItemAt(mpSearchResults->CurrentSelection());
    	if (item!=NULL) {
    		BEntryItem *entry=dynamic_cast<BEntryItem *>(item);
    		if (entry!=NULL) {
    			entry_ref *ref=entry->Ref();
    			be_roster->Launch(ref);
    		}
    	}	
    	break;
    }

    // Let the original BWindow handle the message. --------------------------
    default: {
      BWindow::MessageReceived(apMessage);
      break;
    }
  }
}

// ===========================================================================
// FrameResized
// ===========================================================================

void 
tWindow::FrameResized(float width, float height)
{
  BWindow::FrameResized(width, height);
  SavePrefs();
}

// ===========================================================================
// FrameMoved
// ===========================================================================

void
tWindow::FrameMoved(BPoint screenPoint)
{
  BWindow::FrameMoved(screenPoint);
  SavePrefs();
}

// ===========================================================================
// LoadPrefs
// ===========================================================================

void
tWindow::LoadPrefs()
{
  //puts("[Tracker Grep] LoadPrefs");
  BFile lFile(PREFS_FILE, B_READ_ONLY);
  if (lFile.InitCheck() == B_OK) {

    Lock();
    if (lFile.Lock() == B_OK) {
      
      tInt32 lValue;
      lFile.ReadAttr("RecurseDirs",B_INT32_TYPE,0,&lValue,sizeof(tInt32));
      mpRecurseDirs->SetValue(lValue);

      if (B_CONTROL_ON == mpRecurseDirs->Value()) {
        mpGrepper->SetRecurseDirs(TRUE);
      } else {
        mpGrepper->SetRecurseDirs(FALSE);
      }
        
      lFile.ReadAttr("RecurseLinks",B_INT32_TYPE,0,&lValue,sizeof(tInt32));
      mpRecurseLinks->SetValue(lValue);

      if (B_CONTROL_ON == mpRecurseLinks->Value()) {
        mpGrepper->SetRecurseLinks(TRUE);
      } else {
        mpGrepper->SetRecurseLinks(FALSE);
      }

      lFile.ReadAttr("ShowContents",B_INT32_TYPE,0,&lValue,sizeof(tInt32));
      mpShowContents->SetValue(lValue);

      BRect lRect;
      lFile.ReadAttr("WindowFrame",B_RECT_TYPE,0,&lRect,sizeof(BRect));
      MoveTo(lRect.left,lRect.top);
      ResizeTo(lRect.Width(),lRect.Height());
      
      lFile.Unlock();
    }
    Unlock();
  }
}

// ===========================================================================
// SavePrefs
// ===========================================================================

void
tWindow::SavePrefs()
{
  //puts("[Tracker Grep] SavePrefs");
  BFile lFile(PREFS_FILE, B_CREATE_FILE | B_ERASE_FILE | B_WRITE_ONLY);
  if (lFile.InitCheck() == B_OK) {

    // Lock the window and the file. (Note that the window is usually already
    // locked, because SavePrefs is called from MessageReceived()).
    Lock();
    if (lFile.Lock() == B_OK) {
    
      tInt32 lValue = mpRecurseDirs->Value();
      lFile.WriteAttr("RecurseDirs",B_INT32_TYPE,0,&lValue,sizeof(tInt32));

      lValue = mpRecurseLinks->Value();
      lFile.WriteAttr("RecurseLinks",B_INT32_TYPE,0,&lValue,sizeof(tInt32));

      lValue = mpShowContents->Value();
      lFile.WriteAttr("ShowContents",B_INT32_TYPE,0,&lValue,sizeof(tInt32));

      BRect lRect = Frame();
      lFile.WriteAttr("WindowFrame",B_RECT_TYPE,0,&lRect,sizeof(BRect));
      
      // Write the file immediately to disk, and then unlock everything.
      lFile.Sync();
      lFile.Unlock();
    }
    Unlock();
  }
}

// ***************************************************************************
// PUBLIC Member Function Definitions: Access
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Inquiry
// ***************************************************************************

// ***************************************************************************
// PROTECTED Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// PRIVATE Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// Helper Function Definitions
// ***************************************************************************

// ***************************************************************************
// Global Function Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Definitions
// ***************************************************************************
